from datetime import datetime, date
from decimal import Decimal
from typing import Any, Optional

from normality.cleaning import remove_unsafe_chars
from normality.encoding import predict_encoding
from normality.encoding import DEFAULT_ENCODING


def _clean_empty(value: str) -> Optional[str]:
    # XXX: is this really a good idea?
    value = value.strip()
    if not len(value):
        return None
    return value


def stringify(
    value: Any, encoding_default: str = DEFAULT_ENCODING, encoding: Optional[str] = None
) -> Optional[str]:
    """Brute-force convert a given object to a string.

    This will attempt an increasingly mean set of conversions to make a given
    object into a unicode string. It is guaranteed to either return unicode or
    None, if all conversions failed (or the value is indeed empty).
    """
    if value is None:
        return None
    if isinstance(value, str):
        return _clean_empty(value)
    if isinstance(value, (date, datetime)):
        return value.isoformat()
    elif isinstance(value, (float, Decimal)):
        return Decimal(value).to_eng_string()
    elif isinstance(value, bytes):
        if encoding is None:
            encoding = predict_encoding(value, default=encoding_default)
        value = value.decode(encoding, "replace")
        value = remove_unsafe_chars(value)
        if value is None:
            return None
        return _clean_empty(value)
    return _clean_empty(str(value))



"""Helper functions for string cleaning.

`normality` includes functions to convert arbitrary Python objects to
strings, transliterate them into the latin alphabet, make slugs for
URLs, or perform the substitution of characters based on unicode
character categories.
"""
from typing import Any, Optional

from normality.cleaning import collapse_spaces, category_replace
from normality.constants import UNICODE_CATEGORIES, WS
from normality.transliteration import latinize_text, ascii_text
from normality.encoding import guess_encoding, guess_file_encoding
from normality.encoding import predict_encoding, predict_file_encoding
from normality.encoding import DEFAULT_ENCODING
from normality.stringify import stringify
from normality.paths import safe_filename
from normality.slugify import slugify
from normality.util import Categories, Encoding

__version__ = "2.4.0"
__all__ = [
    "collapse_spaces",
    "category_replace",
    "safe_filename",
    "normalize",
    "stringify",
    "slugify",
    "guess_encoding",
    "guess_file_encoding",
    "predict_encoding",
    "predict_file_encoding",
    "latinize_text",
    "ascii_text",
    "WS",
    "UNICODE_CATEGORIES",
]


def normalize(
    value: Any,
    lowercase: bool = True,
    collapse: bool = True,
    latinize: bool = False,
    ascii: bool = False,
    encoding_default: Encoding = DEFAULT_ENCODING,
    encoding: Optional[str] = None,
    replace_categories: Categories = UNICODE_CATEGORIES,
) -> Optional[str]:
    """The main normalization function for text.

    This will take a string and apply a set of transformations to it so
    that it can be processed more easily afterwards. Arguments:

    * ``lowercase``: not very mysterious.
    * ``collapse``: replace multiple whitespace-like characters with a
      single whitespace. This is especially useful with category replacement
      which can lead to a lot of whitespace.
    * ``decompose``: apply a unicode normalization (NFKD) to separate
      simple characters and their diacritics.
    * ``replace_categories``: This will perform a replacement of whole
      classes of unicode characters (e.g. symbols, marks, numbers) with a
      given character. It is used to replace any non-text elements of the
      input string.
    """
    text = stringify(value, encoding_default=encoding_default, encoding=encoding)
    if text is None:
        return None

    if lowercase:
        # Yeah I made a Python package for this.
        text = text.lower()

    if ascii:
        # A stricter form of transliteration that leaves only ASCII
        # characters.
        text = ascii_text(text)
    elif latinize:
        # Perform unicode-based transliteration, e.g. of cyricllic
        # or CJK scripts into latin.
        text = latinize_text(text)

    # Perform unicode category-based character replacement. This is
    # used to filter out whole classes of characters, such as symbols,
    # punctuation, or whitespace-like characters.
    if replace_categories is not None:
        text = category_replace(text, replace_categories)

    if collapse:
        # Remove consecutive whitespace.
        text = collapse_spaces(text)
    return text



import codecs
import chardet
import warnings
from charset_normalizer import from_bytes, CharsetMatches
from typing import Any, BinaryIO, TYPE_CHECKING
from normality.util import Encoding

if TYPE_CHECKING:
    from charset_normalizer import CharsetMatches

DEFAULT_ENCODING = "utf-8"


def normalize_encoding(encoding: str, default: Encoding = DEFAULT_ENCODING) -> str:
    """Normalize the encoding name, replace ASCII w/ UTF-8."""
    warnings.warn(
        "normalize_encoding is now deprecated. Use tidy_encoding instead",
        DeprecationWarning,
    )
    return tidy_encoding(encoding, default)


def tidy_encoding(encoding: str, default: Encoding = DEFAULT_ENCODING) -> str:
    """Normalize the encoding name, replace ASCII w/ UTF-8."""
    if encoding is None:
        return default
    encoding = encoding.strip()
    if encoding.lower() in ["", "ascii"]:
        return default
    try:
        codec = codecs.lookup(encoding)
        return codec.name
    except LookupError:
        return default


def normalize_result(
    result: Any, default: Encoding, threshold: float = 0.2
) -> Encoding:
    """Interpret a chardet result."""
    warnings.warn(
        "normalize_result is now deprecated. Use tidy_result instead",
        DeprecationWarning,
    )
    if result is None:
        return default
    confidence: float = result.get("confidence")
    if confidence is None:
        return default
    if float(confidence) < threshold:
        return default
    encoding: Encoding = result.get("encoding")
    if encoding is None:
        return default
    return normalize_encoding(encoding, default=default)


def tidy_result(result: CharsetMatches, default: Encoding) -> Encoding:
    """Interpret a chardet result."""
    res = result.best()
    if res is None:
        return default
    encoding: Encoding = res.encoding
    if encoding is None:
        return default

    return tidy_encoding(encoding, default=default)


def guess_encoding(text: bytes, default: Encoding = DEFAULT_ENCODING) -> Encoding:
    """Guess string encoding.

    Given a piece of text, apply character encoding detection to
    guess the appropriate encoding of the text.
    """
    warnings.warn(
        "guess_encoding is now deprecated. Use predict_encoding instead",
        DeprecationWarning,
    )
    return predict_encoding(text, default=default)


def predict_encoding(text: bytes, default: Encoding = DEFAULT_ENCODING) -> Encoding:
    """Guess string encoding.

    Given a piece of text, apply character encoding detection to
    guess the appropriate encoding of the text.
    """
    result = from_bytes(text, explain=False)
    return tidy_result(result, default=default)


def guess_file_encoding(fh: BinaryIO, default: Encoding = DEFAULT_ENCODING) -> Encoding:
    """Guess encoding from a file handle."""
    warnings.warn(
        "guess_encoding is now deprecated. Use predict_encoding instead",
        DeprecationWarning,
    )
    start = fh.tell()
    detector = chardet.UniversalDetector()
    while True:
        data = fh.read(1024 * 10)
        if not data:
            detector.close()
            break
        detector.feed(data)
        if detector.done:
            break

    fh.seek(start)
    return normalize_result(detector.result, default=default)


def predict_file_encoding(
    fh: BinaryIO, default: Encoding = DEFAULT_ENCODING
) -> Encoding:
    """Guess encoding from a file handle."""
    start = fh.tell()
    result: CharsetMatches = CharsetMatches()

    while True:
        data = fh.read(1024 * 10)
        if not data:
            break

        result = from_bytes(data, explain=False)
        if result:
            break

    fh.seek(start)
    return tidy_result(result, default=default)



import re
import unicodedata
from typing import Any, Optional

from normality.constants import UNICODE_CATEGORIES, CONTROL_CODES, WS
from normality.util import Categories, is_text

COLLAPSE_RE = re.compile(r"\s+", re.U)
BOM_RE = re.compile("^\ufeff", re.U)
UNSAFE_RE = re.compile(r"^\ufeff|[\x00-\x08\x0b-\x0c\x0e-\x1f\x7f\x80-\x9f]|\u2028")
QUOTES_RE = re.compile(r'^["\'](.*)["\']$')


def decompose_nfkd(text: Any) -> Optional[str]:
    """Perform unicode compatibility decomposition.

    This will replace some non-standard value representations in unicode and
    normalise them, while also separating characters and their diacritics into
    two separate codepoints.
    """
    if not is_text(text):
        return None
    return unicodedata.normalize("NFKD", text)


def compose_nfc(text: Any) -> Optional[str]:
    """Perform unicode composition."""
    if not is_text(text):
        return None
    return unicodedata.normalize("NFC", text)


def compose_nfkc(text: Any) -> Optional[str]:
    """Perform unicode composition."""
    if not is_text(text):
        return None
    return unicodedata.normalize("NFKC", text)


def strip_quotes(text: Any) -> Optional[str]:
    """Remove double or single quotes surrounding a string."""
    if not is_text(text):
        return None
    return QUOTES_RE.sub("\\1", text)


def category_replace(
    text: Any, replacements: Categories = UNICODE_CATEGORIES
) -> Optional[str]:
    """Remove characters from a string based on unicode classes.

    This is a method for removing non-text characters (such as punctuation,
    whitespace, marks and diacritics) from a piece of text by class, rather
    than specifying them individually.
    """
    text = decompose_nfkd(text)
    if not is_text(text):
        return None
    characters = []
    for character in text:
        cat = unicodedata.category(character)
        replacement = replacements.get(cat, character)
        if replacement is not None:
            characters.append(replacement)
    return "".join(characters)


def remove_control_chars(text: Any) -> Optional[str]:
    """Remove just the control codes from a piece of text."""
    return category_replace(text, replacements=CONTROL_CODES)


def remove_unsafe_chars(text: Any) -> Optional[str]:
    """Remove unsafe unicode characters from a piece of text."""
    if not is_text(text):
        return None
    return UNSAFE_RE.sub("", text)


def remove_byte_order_mark(text: Any) -> Optional[str]:
    """Remove a BOM from the beginning of the text."""
    if not is_text(text):
        return None
    return BOM_RE.sub("", text)


def collapse_spaces(text: Any) -> Optional[str]:
    """Remove newlines, tabs and multiple spaces with single spaces."""
    if not is_text(text):
        return None
    return COLLAPSE_RE.sub(WS, text).strip(WS)



import os
from typing import Optional
from banal import decode_path
from normality.stringify import stringify
from normality.cleaning import collapse_spaces, category_replace
from normality.constants import UNICODE_CATEGORIES, WS
from normality.transliteration import ascii_text

MAX_LENGTH = 254


def _safe_name(file_name: Optional[str], sep: str) -> Optional[str]:
    """Convert the file name to ASCII and normalize the string."""
    file_name = stringify(file_name)
    if file_name is None:
        return None
    file_name = ascii_text(file_name)
    file_name = category_replace(file_name, UNICODE_CATEGORIES)
    file_name = collapse_spaces(file_name)
    if file_name is None or not len(file_name):
        return None
    return file_name.replace(WS, sep)


def safe_filename(
    file_name: Optional[str],
    sep: str = "_",
    default: Optional[str] = None,
    extension: Optional[str] = None,
) -> Optional[str]:
    """Create a secure filename for plain file system storage."""
    if file_name is None:
        return decode_path(default)

    file_name = decode_path(file_name)
    if file_name is None:
        return None
    file_name = os.path.basename(file_name)
    file_name, _extension = os.path.splitext(file_name)
    file_name = _safe_name(file_name, sep=sep)
    if file_name is None:
        return decode_path(default)
    file_name = file_name[:MAX_LENGTH]
    extension = _safe_name(extension or _extension, sep=sep)
    if extension is not None:
        file_name = ".".join((file_name, extension))
        file_name = file_name[:MAX_LENGTH]
    return file_name



import string
from typing import Any, Optional

from normality.cleaning import collapse_spaces, category_replace
from normality.constants import SLUG_CATEGORIES, WS
from normality.transliteration import latinize_text
from normality.stringify import stringify

VALID_CHARS = string.ascii_lowercase + string.digits + WS


def slugify(value: Any, sep: str = "-") -> Optional[str]:
    """A simple slug generator. Slugs are pure ASCII lowercase strings
    that can be used in URLs an other places where a name has to be
    machine-safe."""
    text = stringify(value)
    if text is None:
        return None
    text = text.replace(sep, WS)
    # run this first because it'll give better results on special
    # characters.
    text = category_replace(text, SLUG_CATEGORIES)
    text = latinize_text(text, ascii=True)
    if text is None:
        return None
    text = text.lower()
    text = "".join([c for c in text if c in VALID_CHARS])
    text = collapse_spaces(text)
    if text is None or len(text) == 0:
        return None
    return text.replace(WS, sep)



import unittest

from normality.paths import MAX_LENGTH, safe_filename


class PathsTest(unittest.TestCase):
    def test_safe_filename(self):
        self.assertEqual(None, safe_filename(None))
        self.assertEqual("test.txt", safe_filename("test.txt"))
        self.assertEqual("test.txt", safe_filename("test .txt"))
        self.assertEqual("test_bla.txt", safe_filename("test bla.txt"))
        self.assertEqual("test_bla.txt", safe_filename("test_bla.txt"))
        self.assertEqual("test_bla.txt", safe_filename("test.bla.txt"))
        self.assertEqual("test.txt", safe_filename("test", extension="txt"))

    def test_long_filename(self):
        long_name = ["long name"] * 100
        long_name = "-".join(long_name)

        shortened = safe_filename(long_name)
        assert len(shortened) <= MAX_LENGTH, shortened

        shortened = safe_filename(long_name, extension="html")
        assert len(shortened) <= MAX_LENGTH, shortened

        shortened = safe_filename("bla", extension=long_name)
        assert len(shortened) <= MAX_LENGTH, shortened

        shortened = safe_filename(long_name, extension=long_name)
        assert len(shortened) <= MAX_LENGTH, shortened



import unittest
from datetime import datetime

from normality import normalize, latinize_text, ascii_text
from normality import (
    stringify,
    slugify,
    guess_encoding,
    guess_file_encoding,
    predict_file_encoding,
    predict_encoding,
)


class NormalityTest(unittest.TestCase):
    def test_empty(self):
        self.assertEqual(None, slugify(None))
        self.assertEqual(None, ascii_text(None))
        self.assertEqual(None, latinize_text(None))
        self.assertEqual(None, normalize(None))
        self.assertEqual(None, normalize(""))
        self.assertEqual(None, normalize(" "))

    def test_petro(self):
        text = u"Порошенко Петро Олексійович"
        self.assertEqual("porosenko-petro-oleksijovic", slugify(text))
        self.assertEqual("Porosenko Petro Oleksijovic", ascii_text(text))
        self.assertEqual(u"Porošenko Petro Oleksíjovič", latinize_text(text))
        self.assertEqual(u"порошенко петро олексіиович", normalize(text))

    def test_ahmad(self):
        text = u"əhməd"
        self.assertEqual("ahmad", ascii_text(text))

    def test_azeri(self):
        text = u"FUAD ALIYEV ƏHMƏD OĞLU"
        self.assertEqual("FUAD ALIYEV AHMAD OGLU", ascii_text(text))

    def test_slugify(self):
        text = u"BABY! camel-is good"
        self.assertEqual("baby-camel-is-good", slugify(text, sep="-"))
        self.assertEqual("tests", slugify("testʼs", sep="-"))
        self.assertEqual("test-s", slugify("test_s", sep="-"))
        self.assertEqual(None, slugify("-", sep="-"))
        self.assertEqual(None, slugify("", sep="-"))
        self.assertEqual(None, slugify("- -", sep="-"))
        self.assertEqual(None, slugify(None, sep="-"))

    def test_georgian(self):
        text = u"ავლაბრის ფონდი"
        self.assertEqual("avlabris pondi", ascii_text(text))

    def test_german(self):
        text = u"Häschen Spaß"
        self.assertEqual("Haschen Spass", ascii_text(text))
        self.assertEqual("haschen-spass", slugify(text, sep="-"))

    def test_stringify(self):
        self.assertEqual(".", stringify(" . "))
        self.assertEqual("5", stringify(5))
        self.assertEqual("0.5", stringify(0.5))

    def test_stringify_datetime(self):
        dt = datetime.utcnow()
        text = stringify(dt)
        self.assertTrue(text.startswith("%s-" % dt.year), text)

    def test_guess_encoding(self):
        text = u"Порошенко Петро Олексійович"
        encoded = text.encode("iso-8859-5")
        out = guess_encoding(encoded)
        self.assertEqual("iso8859-5", out)

    def test_predict_encoding(self):
        text = u"Порошенко Петро Олексійович"
        encoded = text.encode("iso-8859-5")
        out = predict_encoding(encoded)
        self.assertEqual("iso8859-5", out)

    def test_guess_file_encoding(self):
        with open("tests/fixtures/utf-16.txt", "rb") as fh:
            out = guess_file_encoding(fh)
            self.assertEqual("utf-16", out)

    def test_predict_file_encoding(self):
        with open("tests/fixtures/utf-16.txt", "rb") as fh:
            out = predict_file_encoding(fh)
            self.assertEqual("utf-16", out)

    def test_petro_iso_encoded(self):
        text = u"Порошенко Петро Олексійович"
        encoded = text.encode("iso8859-5")
        out = stringify(encoded)
        self.assertEqual(text, out)

    def test_petro_utf16_encoded(self):
        text = u"Порошенко Петро Олексійович"
        encoded = text.encode("utf-16")
        out = stringify(encoded)
        self.assertEqual(text, out)



"""
Transliterate the given text to the latin script.

This attempts to convert a given text to latin script using the
closest match of characters vis a vis the original script.

Transliteration requires an extensive unicode mapping. Since all
Python implementations are either GPL-licensed (and thus more
restrictive than this library) or come with a massive C code
dependency, this module requires neither but will use a package
if it is installed.
"""
import warnings
from typing import cast, Optional, Callable
from functools import lru_cache

from normality.cleaning import compose_nfkc
from normality.util import is_text

# Transform to latin, separate accents, decompose, remove
# symbols, compose, push to ASCII
ASCII_SCRIPT = "Any-Latin; NFKD; [:Nonspacing Mark:] Remove; Accents-Any; [:Symbol:] Remove; [:Nonspacing Mark:] Remove; Latin-ASCII"  # noqa
# nb. 2021-11-05 Accents-Any is now followed with another nonspacing mark remover.
# This script is becoming a bit silly, there has to be a nicer way to do this?


class ICUWarning(UnicodeWarning):
    pass


@lru_cache(maxsize=2**16)
def latinize_text(text: Optional[str], ascii: bool = False) -> Optional[str]:
    """Transliterate the given text to the latin script.

    This attempts to convert a given text to latin script using the
    closest match of characters vis a vis the original script.
    """
    if text is None or not is_text(text) or not len(text):
        return text

    if ascii:
        if not hasattr(latinize_text, "_ascii"):
            latinize_text._ascii = make_trans(ASCII_SCRIPT)  # type: ignore
        return latinize_text._ascii(text)  # type: ignore

    if not hasattr(latinize_text, "_tr"):
        latinize_text._tr = make_trans("Any-Latin")  # type: ignore
    return latinize_text._tr(text)  # type: ignore


def ascii_text(text: Optional[str]) -> Optional[str]:
    """Transliterate the given text and make sure it ends up as ASCII."""
    text = latinize_text(text, ascii=True)
    if text is None or not is_text(text):
        return None
    return text.encode("ascii", "replace").decode("ascii")


def make_trans(script: str) -> Callable[[str], Optional[str]]:
    try:
        from icu import Transliterator  # type: ignore

        inst = Transliterator.createInstance(script)
        return cast(Callable[[str], str], inst.transliterate)
    except ImportError:
        from text_unidecode import unidecode  # type: ignore

        warnings.warn(
            "Install 'pyicu' for better text transliteration.", ICUWarning, stacklevel=4
        )  # noqa

        def transliterate(text: str) -> Optional[str]:
            clean = compose_nfkc(text)
            if clean is None:
                return None
            return cast(Optional[str], unidecode(clean))

        return transliterate



from normality.util import Categories

# https://en.wikipedia.org/wiki/Cyrillic_script_in_Unicode
# Cyrillic: U+0400–U+04FF, 256 characters
# Cyrillic Supplement: U+0500–U+052F, 48 characters
# Cyrillic Extended-A: U+2DE0–U+2DFF, 32 characters
# Cyrillic Extended-B: U+A640–U+A69F, 96 characters
# Cyrillic Extended-C: U+1C80–U+1C8F, 9 characters
# Phonetic Extensions: U+1D2B, U+1D78, 2 Cyrillic characters
# Combining Half Marks: U+FE2E–U+FE2F, 2 Cyrillic characters


WS: str = " "

# Unicode character classes, see:
# http://www.fileformat.info/info/unicode/category/index.htm
# https://en.wikipedia.org/wiki/Unicode_character_property
# http://www.unicode.org/charts/beta/script/
UNICODE_CATEGORIES: Categories = {
    "Cc": WS,
    "Cf": None,
    "Cs": None,
    "Co": None,
    "Cn": None,
    "Lm": None,
    "Mn": None,
    "Mc": WS,
    "Me": None,
    "No": None,
    "Zs": WS,
    "Zl": WS,
    "Zp": WS,
    "Pc": WS,  # TODO: figure out if this wants to be None
    "Pd": WS,
    "Ps": WS,
    "Pe": WS,
    "Pi": WS,
    "Pf": WS,
    "Po": WS,
    "Sm": WS,
    "Sc": None,
    "Sk": None,
    "So": WS,
    "Zs": WS,
    "Zl": WS,
    "Zp": WS,
}

SLUG_CATEGORIES: Categories = {
    "Cc": None,
    "Cf": None,
    "Cs": None,
    "Co": None,
    "Cn": None,
    # "Lm": None,
    # "Mn": None,
    "Mc": WS,
    "Me": None,
    "No": None,
    "Zs": WS,
    "Zl": WS,
    "Zp": WS,
    "Pc": WS,
    "Pd": WS,
    "Ps": WS,
    "Pe": WS,
    "Pi": WS,
    "Pf": WS,
    "Po": WS,
    "Sm": WS,
    "Sc": None,
    "Sk": None,
    "So": WS,
    "Zs": WS,
    "Zl": WS,
    "Zp": WS,
}


CONTROL_CODES: Categories = {"Cc": WS, "Cf": WS, "Cs": WS, "Co": WS, "Cn": WS, "Zl": WS}



# Given the whole thing is a utility package, this is really meta.
from typing import Any, Dict, Optional

Categories = Dict[str, Optional[str]]
Encoding = str


def is_text(data: Any) -> bool:
    return isinstance(data, str)
